# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-


""" MIDI channel events.

See also `mode_events` and `voice_events`.

"""

## @namespace midi.channel_events

# There is no `decoded` method here, there is one is `voice_events` and
# another one in `mode_events`, instead.

from enum import IntEnum


# Constants
# ============================================================================

# Ranges
# ----------------------------------------------------------------------------

N_CHANNELS = 16

STATUS_FIRST = 0x80
STATUS_LAST = 0xEF

assert STATUS_FIRST <= STATUS_LAST
# `STATUS_FIRST`..`STATUS_LAST` makes‑up a range.

MODE_CONTROLLER_FIRST = 120


# Statuses
# ----------------------------------------------------------------------------

class Statuses(IntEnum):

    """ Mode or voice channel event statuses with channel set to zero. """

    NOTE_OFF = 0x80
    NOTE_ON = 0x90
    KEY_PRESSURE = 0xA0  # Alias Polyphonic Key Pressure or Aftertouch.
    CONTROL_CHANGE = 0xB0
    PROGRAM_CHANGE = 0xC0
    CHANNEL_PRESSURE = 0xD0  # Alias Channel Pressure or Aftertouch.
    PITCH_BEND = 0xE0  # Alias Pitch Wheel Change or Pitch Bend.


assert len({x.value for x in Statuses}) == len(Statuses)
# `Statuses` contains all unique items.

assert all(
    x >= STATUS_FIRST and
    x <= STATUS_LAST and
    (x % N_CHANNELS) == 0
    for x in Statuses)
# `Statuses` are in `STATUS_FIRST`..`STATUS_LAST` with zeroed channel.

assert all(
    (x % N_CHANNELS) != 0 or
    x in Statuses.__members__.values()
    for x in range(STATUS_FIRST, STATUS_LAST + 1))
# `Statuses` covers all in `STATUS_FIRST`..`STATUS_LAST` with zeroed channel.


# Data sizes
# ----------------------------------------------------------------------------

DATA_SIZES = {
    Statuses.NOTE_OFF: 2,
    Statuses.NOTE_ON: 2,
    Statuses.KEY_PRESSURE: 2,
    Statuses.CONTROL_CHANGE: 2,
    Statuses.PROGRAM_CHANGE: 1,
    Statuses.CHANNEL_PRESSURE: 1,
    Statuses.PITCH_BEND: 2,
}

assert {x for x in DATA_SIZES} == {x for x in Statuses}
# `DATA_SIZES` is a total function of `Statuses`.


# Helper
# ============================================================================

def status_with_zeroed_channel(status):
    """ Status with channel set to zero, to be used as a key. """
    assert status >= STATUS_FIRST and status <= STATUS_LAST
    return (status // N_CHANNELS) * N_CHANNELS


def channel(status):
    """ Channel of status. """
    assert status >= STATUS_FIRST and status <= STATUS_LAST
    return status % N_CHANNELS


# Methods
# ============================================================================

def data_size(status):
    """ Data size for status. """
    assert status >= STATUS_FIRST and status <= STATUS_LAST
    key = status_with_zeroed_channel(status)
    return DATA_SIZES[key]


def is_mode_event(status, data):
    """ If it's to be a `ModeEvent`. """
    assert status >= STATUS_FIRST and status <= STATUS_LAST
    return (
        len(data) == DATA_SIZES[Statuses.CONTROL_CHANGE] and
        status_with_zeroed_channel(status) == Statuses.CONTROL_CHANGE and
        data[0] >= MODE_CONTROLLER_FIRST)
