# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" MIDI Controllers.

"""

## @namespace midi.controller_events

from collections import namedtuple
from enum import (Enum, IntEnum)

from . import voice_events


# Controller Events
# ============================================================================

# Channels and controller indexes, are zero based.
# “Not for setting synthesizer parameters such as VCF cut-off, envelope decay,
# etc”.

Slider7Bits = namedtuple("Slider7Bits", ["channel", "index", "value"])
# `index` may be one of `CoarseSliders` named constants.
# `index` is not the controller number, it's a per kind index.

Command = namedtuple("Command", ["channel", "index"])
# `index` may be one of `Commands` named constants.
# `index` is not the controller number, it's a per kind index.

Slider14Bits = namedtuple("Slider14Bits", ["channel", "index", "msb", "value"])
# `index` may be one of `FineSliders` named constants.
# `index` is not the controller number, it's a per kind index.

Selector = namedtuple("Selector", ["channel", "index", "msb", "value"])
# `index` may be one of `Selectors` named constants.
# `index` is not the controller number, it's a per kind index.

Switch = namedtuple("Switch", ["channel", "index", "on"])
# `index` may be one of `Switches` named constants.
# `index` is not the controller number, it's a per kind index.


# MIDI Defined Controller Numbers (decimal “n” in “CC#n”)
# ============================================================================

FIRST_CONTROLLER = 0
LAST_CONTROLLER = 119
# 120-127 are handled elsewhere as channel mode messages.


class Controllers(IntEnum):

    """ MIDI controllers ID, alias CC#n, where “n” is the ID. """

    # GP is an abbreviation for General Purpose
    # NRPN is an abbreviation for Non‑Registered Parameter Number
    # RPN is an abbreviation for Registered Parameter Number

    # 0-31: 7 MSBits of 14 bits
    # These are the so called “continuous controllers” or “sliders”.
    # MSB is to be sent/received first, LSB is optional, but may come alone
    # too. Do setting MSB resets LSB?
    BANK_MSB = 0  # Alias Bank Select
    MODULATION_MSB = 1  # Alias Modulation Wheel
    BREATH_MSB = 2  # Alias Breath Control
    FOOT_CONTROLLER_MSB = 4
    PORTAMENTO_TIME_MSB = 5
    DATA_ENTRY_MSB = 6  # Related to RPN/NRPN
    CHANNEL_VOLUME_MSB = 7
    BALANCE_MSB = 8
    PAN_MSB = 10
    EXPRESSION_MSB = 11
    EFFECT_1_MSB = 12
    EFFECT_2_MSB = 13
    GP_CONTROLLER_1_MSB = 16
    GP_CONTROLLER_2_MSB = 17
    GP_CONTROLLER_3_MSB = 18
    GP_CONTROLLER_4_MSB = 19

    # 32-63: 7 LSBits of 14 bits
    # LSB of 0-31.
    BANK_LSB = 32  # Alias Bank Select
    MODULATION_LSB = 33  # Alias Modulation Wheel
    BREATH_LSB = 34  # Alias Breath Control
    FOOT_CONTROLLER_LSB = 36
    PORTAMENTO_TIME_LSB = 37
    DATA_ENTRY_LSB = 38  # Related to RPN/NRPN
    CHANNEL_VOLUME_LSB = 39
    BALANCE_LSB = 40
    PAN_LSB = 42
    EXPRESSION_LSB = 43
    EFFECT_1_LSB = 44
    EFFECT_2_LSB = 45
    GP_CONTROLLER_1_LSB = 48
    GP_CONTROLLER_2_LSB = 49
    GP_CONTROLLER_3_LSB = 50
    GP_CONTROLLER_4_LSB = 51

    # 64-69: 0-63 => Off, 64-127 => On
    SUSTAIN = 64  # Alias Damper or Hold Pedal
    PORTAMENTO = 65
    SUSTENUTO = 66
    SOFT = 67  # Alias Soft Pedal.
    LEGATO = 68  # Alias Legato Swithc or Legato Pedal.
    HOLD = 69  # Alias Hold 2 Pedal.

    # 70-95: 7 bits
    SOUND_VARIATION_LSB = 70
    TIMBRE_LSB = 71
    RELEASE_TIME_LSB = 72
    ATTACK_TIME_LSB = 73
    BRIGHTNESS_LSB = 74
    SOUND_CONTROLLER_6_LSB = 75
    SOUND_CONTROLLER_7_LSB = 76
    SOUND_CONTROLLER_8_LSB = 77
    SOUND_CONTROLLER_9_LSB = 78
    SOUND_CONTROLLER_10_LSB = 79
    SOUND_CONTROLLER_11_LSB = 80
    SOUND_CONTROLLER_12_LSB = 81
    SOUND_CONTROLLER_13_LSB = 82
    SOUND_CONTROLLER_14_LSB = 83
    SOUND_CONTROLLER_15_LSB = 84
    SOUND_CONTROLLER_16_LSB = 85
    SOUND_CONTROLLER_17_LSB = 86
    SOUND_CONTROLLER_18_LSB = 87
    SOUND_CONTROLLER_19_LSB = 88
    SOUND_CONTROLLER_20_LSB = 89
    SOUND_CONTROLLER_21_LSB = 90
    EFFECTS_LSB = 91
    TREMULO_LSB = 92
    CHORUS_LSB = 93
    DETUNE_LSB = 94  # Alias Celeste
    PHASER_LSB = 95

    # 96-97: inc/dec
    DATA_ENTRY_INC = 96  # Related to RPN/NRPN
    DATA_ENTRY_DEC = 97  # Related to RPN/NRPN

    # 98-119: LSB/MSB
    NRPN_LSB = 98
    NRPN_MSB = 99
    RPN_LSB = 100
    RPN_MSB = 101


assert all(
    x.value >= FIRST_CONTROLLER and
    x.value <= LAST_CONTROLLER
    for x in Controllers)

assert len({x.value for x in Controllers}) == len(Controllers)


# Mapping from controller numbers to kind and per‑kind indexes.
# ============================================================================

# Kinds for `Mapping` type below.
class Kinds(Enum):

    """ Controller kinds. """

    SLIDER_7_BITS = 0
    COMMAND = 1
    SLIDER_14_BITS_LSB = 2
    SLIDER_14_BITS_MSB = 3
    OMNI_CHANNELS = 4
    SELECTOR_LSB = 5
    SELECTOR_MSB = 6
    SWITCH = 7


Mapping = namedtuple("Mapping", ["kind", "index"])
# `index` distinguishes controls of the same `kind`.
# `index` starts at zero and are consecutives (checked by assertions).


MAPPINGS = [
    Mapping(kind=Kinds.SELECTOR_MSB, index=0),  # 0: BANK_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=0),  # 1: MODULATION_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=1),  # 2: BREATH_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=2),  # 3:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=3),  # 4: FOOT_CONTROLLER_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=4),  # 5: PORTAMENTO_TIME_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=5),  # 6: DATA_ENTRY_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=6),  # 7: CHANNEL_VOLUME_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=7),  # 8: BALANCE_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=8),  # 9:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=9),  # 10: PAN_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=10),  # 11: EXPRESSION_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=11),  # 12: EFFECT_1_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=12),  # 13: EFFECT_2_MSB
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=13),  # 14:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=14),  # 15:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=15),  # 16: GP_CONTROLLER_1_M
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=16),  # 17: GP_CONTROLLER_2_M
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=17),  # 18: GP_CONTROLLER_3_M
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=18),  # 19: GP_CONTROLLER_4_M
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=19),  # 20:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=20),  # 21:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=21),  # 22:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=22),  # 23:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=23),  # 24:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=24),  # 25:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=25),  # 26:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=26),  # 27:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=27),  # 28:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=28),  # 29:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=29),  # 30:
    Mapping(kind=Kinds.SLIDER_14_BITS_MSB, index=30),  # 31:
    Mapping(kind=Kinds.SELECTOR_LSB, index=0),  # 32: BANK_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=0),  # 33: MODULATION_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=1),  # 34: BREATH_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=2),  # 35:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=3),  # 36: FOOT_CONTROLLER_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=4),  # 37: PORTAMENTO_TIME_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=5),  # 38: DATA_ENTRY_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=6),  # 39: CHANNEL_VOLUME_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=7),  # 40: BALANCE_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=8),  # 41:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=9),  # 42: PAN_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=10),  # 43: EXPRESSION_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=11),  # 44: EFFECT_1_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=12),  # 45: EFFECT_2_LSB
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=13),  # 46:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=14),  # 47:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=15),  # 48: GP_CONTROLLER_1_L
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=16),  # 49: GP_CONTROLLER_2_L
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=17),  # 50: GP_CONTROLLER_3_L
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=18),  # 51: GP_CONTROLLER_4_L
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=19),  # 52:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=20),  # 53:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=21),  # 54:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=22),  # 55:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=23),  # 56:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=24),  # 57:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=25),  # 58:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=26),  # 59:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=27),  # 60:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=28),  # 61:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=29),  # 62:
    Mapping(kind=Kinds.SLIDER_14_BITS_LSB, index=30),  # 63:
    Mapping(kind=Kinds.SWITCH, index=0),  # 64: SUSTAIN
    Mapping(kind=Kinds.SWITCH, index=1),  # 65: PORTAMENTO
    Mapping(kind=Kinds.SWITCH, index=2),  # 66: SUSTENUTO
    Mapping(kind=Kinds.SWITCH, index=3),  # 67: SOFT
    Mapping(kind=Kinds.SWITCH, index=4),  # 68: LEGATO
    Mapping(kind=Kinds.SWITCH, index=5),  # 69: HOLD
    Mapping(kind=Kinds.SLIDER_7_BITS, index=0),  # 70: SOUND_VARIATION_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=1),  # 71: TIMBRE_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=2),  # 72: RELEASE_TIME_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=3),  # 73: ATTACK_TIME_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=4),  # 74: BRIGHTNESS_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=5),  # 75: SOUND_CONTROLLER_6_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=6),  # 76: SOUND_CONTROLLER_7_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=7),  # 77: SOUND_CONTROLLER_8_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=8),  # 78: SOUND_CONTROLLER_9_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=9),  # 79: SOUND_CONTROLLER_10_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=10),  # 80: …_11_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=11),  # 81: …_12_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=12),  # 82: …_13_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=13),  # 83: …_14_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=14),  # 84: …_15_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=15),  # 85: …_16_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=16),  # 86: …_17_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=17),  # 87: …_18_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=18),  # 88: …_19_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=19),  # 89: …_20_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=20),  # 90: …_21_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=21),  # 91: EFFECTS_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=22),  # 92: TREMULO_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=23),  # 93: CHORUS_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=24),  # 94: DETUNE_LSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=25),  # 95: PHASER_LSB
    Mapping(kind=Kinds.COMMAND, index=0),  # 96: DATA_ENTRY_INC
    Mapping(kind=Kinds.COMMAND, index=1),  # 97: DATA_ENTRY_DEC
    Mapping(kind=Kinds.SELECTOR_LSB, index=1),  # 98: NRPN_LSB
    Mapping(kind=Kinds.SELECTOR_MSB, index=1),  # 99: NRPN_MSB
    Mapping(kind=Kinds.SELECTOR_LSB, index=2),  # 100: RPN_LSB
    Mapping(kind=Kinds.SELECTOR_MSB, index=2),  # 101: RPN_MSB
    Mapping(kind=Kinds.SLIDER_7_BITS, index=26),  # 102:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=27),  # 103:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=28),  # 104:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=29),  # 105:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=30),  # 106:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=31),  # 107:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=32),  # 108:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=33),  # 109:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=34),  # 110:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=35),  # 111:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=36),  # 112:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=37),  # 113:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=38),  # 114:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=39),  # 115:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=40),  # 116:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=41),  # 117:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=42),  # 118:
    Mapping(kind=Kinds.SLIDER_7_BITS, index=43),  # 119:

    # 120-127 are handled elsewhere as channel mode messages.
]


def is_sequence(l):  # pylint: disable=invalid-name
    """ For an assertion. """
    return all(l[i] == l[i - 1] + 1 for i in range(1, len(l)))


assert len(MAPPINGS) >= len(Controllers)
assert len(MAPPINGS) == 120
KINDS = {x.kind for x in MAPPINGS}
INDEXES = {k: [x.index for x in MAPPINGS if x.kind == k] for k in KINDS}
assert all(len(l) >= 1 and l[0] == 0 for l in INDEXES.values())
assert all(is_sequence(l) for l in INDEXES.values())
assert (
    len(INDEXES[Kinds.SLIDER_14_BITS_LSB]) ==
    len(INDEXES[Kinds.SLIDER_14_BITS_MSB]))
assert len(INDEXES[Kinds.SELECTOR_LSB]) == len(INDEXES[Kinds.SELECTOR_MSB])


# Per Kind Named Indexes
# ============================================================================

# Commands
# ----------------------------------------------------------------------------

def index2(ctrl_id):
    """ Get index checking assertions. """
    # Should be a static class method, but can't initializes a class constant.
    assert MAPPINGS[ctrl_id].kind == Kinds.COMMAND
    return MAPPINGS[ctrl_id].index


class Commands(IntEnum):

    """ Command Indexes. """

    DATA_ENTRY_INC = index2(Controllers.DATA_ENTRY_INC)
    DATA_ENTRY_DEC = index2(Controllers.DATA_ENTRY_DEC)


N_COMMANDS = len(
    {x for x in range(0, len(MAPPINGS))
     if MAPPINGS[x].kind == Kinds.COMMAND})


# Selectors
# ----------------------------------------------------------------------------

def index4(id_lsb, id_msb):
    """ Get index checking assertions. """
    # Should be a static class method, but can't initializes a class constant.
    assert MAPPINGS[id_lsb].index == MAPPINGS[id_msb].index
    assert MAPPINGS[id_lsb].kind == Kinds.SELECTOR_LSB
    assert MAPPINGS[id_msb].kind == Kinds.SELECTOR_MSB
    return MAPPINGS[id_lsb].index


class Selectors(IntEnum):  # pylint: disable=too-few-public-methods

    """ Selectors Indexes. """

    BANK = index4(Controllers.BANK_LSB, Controllers.BANK_MSB)
    NRPN = index4(Controllers.NRPN_LSB, Controllers.NRPN_MSB)
    RPN = index4(Controllers.RPN_LSB, Controllers.RPN_MSB)  # See `rpns`.


N_SELECTORS_LSB = len(
    {x for x in range(0, len(MAPPINGS))
     if MAPPINGS[x].kind == Kinds.SELECTOR_LSB})
N_SELECTORS_MSB = len(
    {x for x in range(0, len(MAPPINGS))
     if MAPPINGS[x].kind == Kinds.SELECTOR_MSB})
assert N_SELECTORS_LSB == N_SELECTORS_MSB
N_SELECTORS = N_SELECTORS_LSB


# 7 Bits Sliders
# ----------------------------------------------------------------------------

def index1(id_lsb):
    """ Get index checking assertions. """
    # Should be a static class method, but can't initializes a class constant.
    assert MAPPINGS[id_lsb].kind == Kinds.SLIDER_7_BITS
    return MAPPINGS[id_lsb].index


class Sliders7Bits(IntEnum):

    """ 7 Bits Slider Indexes. """

    SOUND_VARIATION = index1(Controllers.SOUND_VARIATION_LSB)
    TIMBRE = index1(Controllers.TIMBRE_LSB)
    RELEASE_TIME = index1(Controllers.RELEASE_TIME_LSB)
    ATTACK_TIME = index1(Controllers.ATTACK_TIME_LSB)
    BRIGHTNESS = index1(Controllers.BRIGHTNESS_LSB)
    SOUND_CONTROLLER_6 = index1(Controllers.SOUND_CONTROLLER_6_LSB)
    SOUND_CONTROLLER_7 = index1(Controllers.SOUND_CONTROLLER_7_LSB)
    SOUND_CONTROLLER_8 = index1(Controllers.SOUND_CONTROLLER_8_LSB)
    SOUND_CONTROLLER_9 = index1(Controllers.SOUND_CONTROLLER_9_LSB)
    SOUND_CONTROLLER_10 = index1(Controllers.SOUND_CONTROLLER_10_LSB)
    SOUND_CONTROLLER_11 = index1(Controllers.SOUND_CONTROLLER_11_LSB)
    SOUND_CONTROLLER_12 = index1(Controllers.SOUND_CONTROLLER_12_LSB)
    SOUND_CONTROLLER_13 = index1(Controllers.SOUND_CONTROLLER_13_LSB)
    SOUND_CONTROLLER_14 = index1(Controllers.SOUND_CONTROLLER_14_LSB)
    SOUND_CONTROLLER_15 = index1(Controllers.SOUND_CONTROLLER_15_LSB)
    SOUND_CONTROLLER_16 = index1(Controllers.SOUND_CONTROLLER_16_LSB)
    SOUND_CONTROLLER_17 = index1(Controllers.SOUND_CONTROLLER_17_LSB)
    SOUND_CONTROLLER_18 = index1(Controllers.SOUND_CONTROLLER_18_LSB)
    SOUND_CONTROLLER_19 = index1(Controllers.SOUND_CONTROLLER_19_LSB)
    SOUND_CONTROLLER_20 = index1(Controllers.SOUND_CONTROLLER_20_LSB)
    SOUND_CONTROLLER_21 = index1(Controllers.SOUND_CONTROLLER_21_LSB)
    EFFECTS = index1(Controllers.EFFECTS_LSB)
    TREMULO = index1(Controllers.TREMULO_LSB)
    CHORUS = index1(Controllers.CHORUS_LSB)
    DETUNE = index1(Controllers.DETUNE_LSB)
    PHASER = index1(Controllers.PHASER_LSB)


N_SLIDERS_7_BITS = len(
    {x for x in range(0, len(MAPPINGS))
     if MAPPINGS[x].kind == Kinds.SLIDER_7_BITS})


# 14 Bits Sliders
# ----------------------------------------------------------------------------

def index3(id_lsb, id_msb):
    """ Get index checking assertions. """
    # Should be a static class method, but can't initializes a class constant.
    assert MAPPINGS[id_lsb].index == MAPPINGS[id_msb].index
    assert MAPPINGS[id_lsb].kind == Kinds.SLIDER_14_BITS_LSB
    assert MAPPINGS[id_msb].kind == Kinds.SLIDER_14_BITS_MSB
    return MAPPINGS[id_lsb].index


class Sliders14Bits(IntEnum):

    """ 14 Bits Sliders Indexes. """

    MODULATION = index3(
        Controllers.MODULATION_LSB,
        Controllers.MODULATION_MSB)
    BREATH = index3(
        Controllers.BREATH_LSB,
        Controllers.BREATH_MSB)
    FOOT_CONTROLLER = index3(
        Controllers.FOOT_CONTROLLER_LSB,
        Controllers.FOOT_CONTROLLER_MSB)
    PORTAMENTO_TIME = index3(
        Controllers.PORTAMENTO_TIME_LSB,
        Controllers.PORTAMENTO_TIME_MSB)
    DATA_ENTRY = index3(
        Controllers.DATA_ENTRY_LSB,
        Controllers.DATA_ENTRY_MSB)
    CHANNEL_VOLUME = index3(
        Controllers.CHANNEL_VOLUME_LSB,
        Controllers.CHANNEL_VOLUME_MSB)
    BALANCE = index3(
        Controllers.BALANCE_LSB,
        Controllers.BALANCE_MSB)
    PAN = index3(
        Controllers.PAN_LSB,
        Controllers.PAN_MSB)
    EXPRESSION = index3(
        Controllers.EXPRESSION_LSB,
        Controllers.EXPRESSION_MSB)
    EFFECT_1 = index3(
        Controllers.EFFECT_1_LSB,
        Controllers.EFFECT_1_MSB)
    EFFECT_2 = index3(
        Controllers.EFFECT_2_LSB,
        Controllers.EFFECT_2_MSB)
    GP_CONTROLLER_1 = index3(
        Controllers.GP_CONTROLLER_1_LSB,
        Controllers.GP_CONTROLLER_1_MSB)
    GP_CONTROLLER_2 = index3(
        Controllers.GP_CONTROLLER_2_LSB,
        Controllers.GP_CONTROLLER_2_MSB)
    GP_CONTROLLER_3 = index3(
        Controllers.GP_CONTROLLER_3_LSB,
        Controllers.GP_CONTROLLER_3_MSB)
    GP_CONTROLLER_4 = index3(
        Controllers.GP_CONTROLLER_4_LSB,
        Controllers.GP_CONTROLLER_4_MSB)


N_SLIDERS_14_BITS_LSB = len(
    {x for x in range(0, len(MAPPINGS))
     if MAPPINGS[x].kind == Kinds.SLIDER_14_BITS_LSB})
N_SLIDERS_14_BITS_MSB = len(
    {x for x in range(0, len(MAPPINGS))
     if MAPPINGS[x].kind == Kinds.SLIDER_14_BITS_MSB})
assert N_SLIDERS_14_BITS_LSB == N_SLIDERS_14_BITS_MSB
N_SLIDERS_14_BITS = N_SLIDERS_14_BITS_LSB


# Switches
# ----------------------------------------------------------------------------

def index5(ctrl_id):
    """ Get index checking assertions. """
    # Should be a static class method, but can't initializes a class constant.
    assert MAPPINGS[ctrl_id].kind == Kinds.SWITCH
    return MAPPINGS[ctrl_id].index


class Switches(IntEnum):

    """ Switch Indexes. """

    SUSTAIN = index5(Controllers.SUSTAIN)
    PORTAMENTO = index5(Controllers.PORTAMENTO)
    SUSTENUTO = index5(Controllers.SUSTENUTO)
    SOFT = index5(Controllers.SOFT)
    LEGATO = index5(Controllers.LEGATO)
    HOLD = index5(Controllers.HOLD)


N_SWITCHES = len(
    {x for x in range(0, len(MAPPINGS))
     if MAPPINGS[x].kind == Kinds.SWITCH})


# Main
# ============================================================================

def decoded(event):
    """ Controller event from channel event. """
    # Event is a channel event, not a raw MIDI event.
    if not isinstance(event, voice_events.ControlChange):
        raise RuntimeError
    if event.controller >= len(MAPPINGS):
        raise RuntimeError

    mapping = MAPPINGS[event.controller]
    kind = mapping.kind
    index = mapping.index
    channel = event.channel
    value = event.value
    if value > 127:
        raise RuntimeError

    if kind == Kinds.SLIDER_7_BITS:
        result = Slider7Bits(
            channel=channel,
            index=index,
            value=value)
    elif kind == Kinds.COMMAND:
        # Ignore value
        result = Command(channel=channel, index=index)
    elif kind in [Kinds.SLIDER_14_BITS_LSB, Kinds.SLIDER_14_BITS_MSB]:
        msb = (kind == Kinds.SLIDER_14_BITS_MSB)
        result = Slider14Bits(
            channel=channel,
            index=index,
            msb=msb,
            value=value)
    elif kind in [Kinds.SELECTOR_LSB, Kinds.SELECTOR_MSB]:
        msb = (kind == Kinds.SELECTOR_MSB)
        result = Selector(
            channel=channel,
            index=index,
            msb=msb,
            value=value)
    elif kind == Kinds.SWITCH:
        switch_on = (value >= 64)
        result = Switch(channel=channel, index=index, on=switch_on)
    else:
        raise RuntimeError

    return result
