# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" Channel event, meta event and system event, names. """

from . import midi

from .channel_events import Statuses as ChannelEvents
from .meta_events import Types as MetaEvents
from .mode_events import Controllers as ModeControllers
from .realtime_events import Statuses as RealtimeStatuses
from .system_events import Statuses as SystemEvents


# Per event class
# ============================================================================

# Meta Events
# ----------------------------------------------------------------------------

# Indexed by type.
META_EVENT_NAMES = {
    MetaEvents.SEQUENCE_NUMBER: "Sequence Number",
    MetaEvents.TEXT_EVENT: "Text Event",
    MetaEvents.COPYRIGHT_NOTICE: "Copyright Notice",
    MetaEvents.SEQUENCE_OR_TRACK_NAME: "Sequence or Track Name",
    MetaEvents.INSTRUMENT_NAME: "Instrument Name",
    MetaEvents.LYRIC: "Lyric",
    MetaEvents.MARKER: "Marker",
    MetaEvents.CUE_POINT: "Cue Point",
    MetaEvents.DEVICE_OR_PORT_NAME: "Device or Port Name",
    MetaEvents.CHANNEL_PREFIX: "Channel Prefix",
    MetaEvents.PORT_PREFIX: "Port Prefix",
    MetaEvents.END_OF_TRACK: "End of Track",
    MetaEvents.TEMPO: "Tempo",
    MetaEvents.SMPTE_OFFSET: "SMPTE Offset",
    MetaEvents.TIME_SIGNATURE: "Time Signature",
    MetaEvents.KEY_SIGNATURE: "Key Signature",
    MetaEvents.SEQUENCER_SPECIFIC_EVENT: "Sequencer Specific Event",
}

assert len(META_EVENT_NAMES) == len(MetaEvents)
assert all(x.value in META_EVENT_NAMES for x in MetaEvents)


def meta_event_name(event):
    """ Meta‑event name. """
    if not isinstance(event, midi.MetaEvent):
        raise RuntimeError
    index = event.type
    if index not in META_EVENT_NAMES:
        raise RuntimeError
    return META_EVENT_NAMES[index]


# Mode Events
# ----------------------------------------------------------------------------

# Indexed by controllers.
MODE_EVENT_NAMES = {
    ModeControllers.ALL_SOUND_OFF: "All Sound Off",
    ModeControllers.RESET_ALL_CONTROLLERS: "Reset All Controllers",
    ModeControllers.LOCAL_CONTROL: "Local Control On/Off",
    ModeControllers.ALL_NOTES_OFF: "All Notes Off",
    ModeControllers.OMNI_OFF: "Omni Off",
    ModeControllers.OMNI_ON: "Omni On",
    ModeControllers.MONO_ON: "Mono On",
    ModeControllers.POLY_ON: "Poly On",
}

assert len(MODE_EVENT_NAMES) == len(ModeControllers)
assert all(x.value in MODE_EVENT_NAMES for x in ModeControllers)


def mode_event_name(event):
    """ Mode‑event name. """
    if not isinstance(event, midi.ModeEvent):
        raise RuntimeError
    if len(event.data) != 2:
        raise RuntimeError
    index = event.data[0]
    if index not in MODE_EVENT_NAMES:
        raise RuntimeError
    return MODE_EVENT_NAMES[index]


# Realtime Events
# ----------------------------------------------------------------------------

# Indexed by status.
REALTIME_EVENT_NAMES = {
    RealtimeStatuses.CLOCK: "Clock",
    RealtimeStatuses.TICK: "Tick",
    RealtimeStatuses.START: "Strart",
    RealtimeStatuses.CONTINUE: "Continue",
    RealtimeStatuses.STOP: "Stop",
    RealtimeStatuses.ACTIVE_SENSE: "Active Sense",
    RealtimeStatuses.RESET: "Reset",
}

assert len(REALTIME_EVENT_NAMES) == len(RealtimeStatuses)
assert all(x.value in REALTIME_EVENT_NAMES for x in RealtimeStatuses)


def realtime_event_name(event):
    """ Realtime‑event name. """
    if not isinstance(event, midi.RealtimeEvent):
        raise RuntimeError
    index = event.status
    if index not in REALTIME_EVENT_NAMES:
        raise RuntimeError
    return REALTIME_EVENT_NAMES[index]


# SysEx Events
# ----------------------------------------------------------------------------

SYSEX_EVENT_NAME = "SysEx"


def sysex_event_name(event):
    """ SysEx‑event name. """
    if not isinstance(event, midi.SysExEvent):
        raise RuntimeError
    return SYSEX_EVENT_NAME


# System Events
# ----------------------------------------------------------------------------

# Indexed by status.
SYSTEM_EVENT_NAMES = {
    SystemEvents.QUARTER_FRAME: "Quarter Frame",
    SystemEvents.SONG_POSITION_POINTER: "Song Position Pointer",
    SystemEvents.SONG_SELECT: "Song Select",
    SystemEvents.TUNE_REQUEST: "Tune Request",
}

assert len(SYSTEM_EVENT_NAMES) == len(SystemEvents)
assert all(x.value in SYSTEM_EVENT_NAMES for x in SystemEvents)


def system_event_name(event):
    """ System event name. """
    if not isinstance(event, midi.SystemEvent):
        raise RuntimeError
    index = event.status
    if index not in SYSTEM_EVENT_NAMES:
        raise RuntimeError
    return SYSTEM_EVENT_NAMES[index]


# Voice Events
# ----------------------------------------------------------------------------

# Indexed by status with zeroed channel.
VOICE_EVENT_NAMES = {
    ChannelEvents.NOTE_OFF: "Note Off",
    ChannelEvents.NOTE_ON: "Note On",
    ChannelEvents.KEY_PRESSURE: "Key Pressure",
    ChannelEvents.CONTROL_CHANGE: "Control Change",
    ChannelEvents.PROGRAM_CHANGE: "Program Change",
    ChannelEvents.CHANNEL_PRESSURE: "Channel Pressure",
    ChannelEvents.PITCH_BEND: "Pitch Bend",
}

assert len(VOICE_EVENT_NAMES) == len(ChannelEvents)
assert all(x.value in VOICE_EVENT_NAMES for x in ChannelEvents)


def voice_event_name(event):
    """ Channel voice event name.

    Does not interpret note‑on with velocity zero as note‑off! It just says
    note‑on, in this case.

    """
    if not isinstance(event, midi.VoiceEvent):
        raise RuntimeError
    index = ((event.status) // 16) * 16
    if index not in VOICE_EVENT_NAMES:
        raise RuntimeError
    return VOICE_EVENT_NAMES[index]


# Main
# ============================================================================

def name(event):
    """ Event name. """
    if isinstance(event, midi.MetaEvent):
        result = meta_event_name(event)
    elif isinstance(event, midi.ModeEvent):
        result = mode_event_name(event)
    elif isinstance(event, midi.RealtimeEvent):
        result = realtime_event_name(event)
    elif isinstance(event, midi.SysExEvent):
        result = sysex_event_name(event)
    elif isinstance(event, midi.SystemEvent):
        result = system_event_name(event)
    elif isinstance(event, midi.VoiceEvent):
        result = voice_event_name(event)
    else:
        raise RuntimeError
    return result
