# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" MIDI Meta‑Events. """

from collections import namedtuple
from enum import IntEnum

import smpte
from . import midi
from . import manufacturers


# BPM = 60,000,000/MicroTempo
# MicrosPerPPQN = MicroTempo/TimeBase
# MicrosPerMIDIClock = MicroTempo/24
# PPQNPerMIDIClock = TimeBase/24
# MicrosPerSubFrame = 1000000 * Frames * SubFrames
# SubFramesPerQuarterNote = MicroTempo/(Frames * SubFrames)
# SubFramesPerPPQN = SubFramesPerQuarterNote/TimeBase
# MicrosPerPPQN = SubFramesPerPPQN * Frames * SubFrames

# Constants
# ============================================================================

FIRST_TYPE = 0x00
LAST_TYPE = 0x7F


class Types(IntEnum):

    """ Meta event constants. """

    SEQUENCE_NUMBER = 0x00  # 16 bits integer
    TEXT_EVENT = 0x01  # variable length text
    COPYRIGHT_NOTICE = 0x02  # variable length text
    SEQUENCE_OR_TRACK_NAME = 0x03  # variable length text
    INSTRUMENT_NAME = 0x04  # variable length text
    LYRIC = 0x05  # variable length text
    MARKER = 0x06  # variable length text
    CUE_POINT = 0x07  # variable length text
    DEVICE_OR_PORT_NAME = 0x09  # variable length text
    CHANNEL_PREFIX = 0x20  # 4 bits channel (or associated track number)
    PORT_PREFIX = 0x21  # 8 bits port (or associated track number)
    END_OF_TRACK = 0x2F  # Syntactical, not seen by the application
    TEMPO = 0x51  # 24 bits microsecond length of a quarter note
    SMPTE_OFFSET = 0x54  # hr mn se fr ff (as 1/100 of a frame), 8 bits each
    TIME_SIGNATURE = 0x58  # nn dd cc bb, 8 bits each
    KEY_SIGNATURE = 0x59  # sf mi, 8 bits each
    SEQUENCER_SPECIFIC_EVENT = 0x7F  # variable length bytes


assert all(x.value >= FIRST_TYPE and x.value <= LAST_TYPE for x in Types)
assert len({x.value for x in Types}) == len(Types)


TEXT_EVENTS = {
    Types.TEXT_EVENT,
    Types.COPYRIGHT_NOTICE,
    Types.SEQUENCE_OR_TRACK_NAME,
    Types.INSTRUMENT_NAME,
    Types.LYRIC,
    Types.MARKER,
    Types.CUE_POINT,
    Types.DEVICE_OR_PORT_NAME,
}

assert len(TEXT_EVENTS) <= len(Types)
assert all(x in {y.value for y in Types} for x in TEXT_EVENTS)


def is_text_event(event):
    """ `True` if a text meta‑event. """
    if not isinstance(event, midi.MetaEvent):
        raise RuntimeError
    return event.type in TEXT_EVENTS


# Types
# ============================================================================

ChannelPrefix = namedtuple("ChannelPrefix", ["value"])
CopyrightNotice = namedtuple("CopyrightNotice", ["value"])
CuePoint = namedtuple("CuePoint", ["value"])
DeviceOrPortName = namedtuple("DeviceOrPortName", ["value"])
InstrumentName = namedtuple("InstrumentName", ["value"])
KeySignature = namedtuple("KeySignature", ["alterations", "major"])
Lyric = namedtuple("Lyric", ["value"])
Marker = namedtuple("Marker", ["value"])
PortPrefix = namedtuple("PortPrefix", ["value"])
SequenceNumber = namedtuple("SequenceNumber", ["value"])
SequenceOrTrackName = namedtuple("SequenceOrTrackName", ["value"])
SequencerEvent = namedtuple("SequencerEvent", ["manufacturer", "data"])
SMPTEOffset = namedtuple("SMPTEOffset", ["rate", "address"])
Tempo = namedtuple("Tempo", ["value"])
Text = namedtuple("Text", ["value"])
UnknownMetaEvent = namedtuple("UnknownMetaEvent", ["event"])
TimeSignature = namedtuple(
    "TimeSignature",
    ["numerator", "denominator", "click_clocks", "note_scale"])


# Common Helper
# ============================================================================

def text(data):
    """ Text from data. Note the result may be zero‑length!

    Aware of zero which may occurs and UTF-8 vs ISO-8859-15 encoding.

    """
    length = 0
    while length < len(data):
        if data[length] == 0:
            break
        length += 1

    try:
        result = data[:length].decode("utf8")
    except UnicodeDecodeError:
        result = data[:length].decode("iso-8859-15")

    return result


# Per Type
# ============================================================================

def decoded_text(data):
    """ `Text`. """
    return Text(value=text(data))


def decoded_copyright_notice(data):
    """ `CopyrightNotice`. """
    return CopyrightNotice(value=text(data))


def decoded_sequence_or_track_name(data):
    """ `SequenceOrTrackName`. """
    return SequenceOrTrackName(value=text(data))


def decoded_instrument_name(data):
    """ `InstrumentName`. """
    return InstrumentName(value=text(data))


def decoded_lyric(data):
    """ `Lyric`. """
    return Lyric(value=text(data))


def decoded_marker(data):
    """ `Marker`. """
    return Marker(value=text(data))


def decoded_cue_point(data):
    """ `CuePoint`. """
    return CuePoint(value=text(data))


def decoded_device_or_port_name(data):
    """ `DeviceOrPortName`. """
    return DeviceOrPortName(value=text(data))


def decoded_sequence_number(data):
    """ Sequence number, 0-65535, from data. """
    if len(data) != 2:
        raise ValueError
    value = midi.be16bits(data[0], data[1])
    return SequenceNumber(value=value)


def decoded_channel_prefix(data):
    """ Channel, 0-15, from data. """
    if len(data) != 1:
        raise ValueError
    value = data[0]
    if value > 15:
        raise ValueError
    return ChannelPrefix(value=value)


def decoded_port_prefix(data):
    """ Port, 0-255, from data. """
    if len(data) != 1:
        raise ValueError
    value = data[0]
    return PortPrefix(value=value)


def decoded_tempo(data):
    """ Quarter note duration in microsecond, 1-16777215, from data. """
    if len(data) != 3:
        raise ValueError
    value = midi.be24bits(data[0], data[1], data[2])
    if value == 0:
        raise ValueError
    return Tempo(value=value)


def decoded_smpte_offset(data):
    """ `SMPTEOffset` from data. """

    if len(data) != 5:
        raise ValueError

    rate_code = (data[0] // 32) % 4  # Ignore bits #7
    hour = data[0] % 32
    minute = data[1] % 64  # Ignore bits #6 and #7
    second = data[2] % 64  # Ignore bits #6 and #7
    frame = data[3] % 32  # Ignore bits #5, #6 and #7
    subframe = data[4]

    if rate_code == 0:
        rate = smpte.RATE_24FPS
    elif rate_code == 1:
        rate = smpte.RATE_25FPS
    elif rate_code == 2:
        rate = smpte.RATE_30FPS_DF
    elif rate_code == 3:
        rate = smpte.RATE_30FPS
    else:
        raise RuntimeError

    if hour > 23:
        raise ValueError
    if minute > 59:
        raise ValueError
    if second > 59:
        raise ValueError
    if frame > 29:
        raise ValueError
    if subframe > 99:
        raise ValueError

    address = smpte.FractionalAddress(
        hour=hour,
        minute=minute,
        second=second,
        frame=frame,
        subframe=subframe)

    return SMPTEOffset(rate=rate, address=address)


def decoded_time_signature(data):
    """ `TimeSignature` from data. """

    if len(data) != 4:
        raise ValueError

    numerator = data[0]
    denominator = 2 ** data[1]
    click_clocks = data[2]
    note_scale = data[3] / 8

    if numerator == 0:
        raise ValueError
    if click_clocks == 0:
        raise ValueError
    if note_scale == 0:
        raise ValueError

    return TimeSignature(
        numerator=numerator,
        denominator=denominator,
        click_clocks=click_clocks,
        note_scale=note_scale)


def decoded_key_signature(data):
    """ `KeySignature` from data. """

    if len(data) != 2:
        raise ValueError

    alterations = data[0]
    minor = data[1]

    if alterations > 7 and alterations < 249:
        raise ValueError
    if minor > 1:
        raise ValueError

    if alterations > 7:
        alterations = alterations - 256

    return KeySignature(alterations=alterations, major=(minor != 1))


def decoded_sequencer_event(data):
    """ `SequencerEvent` from data. """

    if not data:  # len(data) == 0
        raise ValueError

    manufacturer_data = manufacturers.decoded_data(data)
    manufacturer = manufacturer_data.manufacturer
    data = manufacturer_data.data
    return SequencerEvent(manufacturer=manufacturer, data=data)


# Main
# ============================================================================

MAPPING = {
    Types.SEQUENCE_NUMBER: decoded_sequence_number,
    Types.TEXT_EVENT: decoded_text,
    Types.COPYRIGHT_NOTICE: decoded_copyright_notice,
    Types.SEQUENCE_OR_TRACK_NAME: decoded_sequence_or_track_name,
    Types.INSTRUMENT_NAME: decoded_instrument_name,
    Types.LYRIC: decoded_lyric,
    Types.MARKER: decoded_marker,
    Types.CUE_POINT: decoded_cue_point,
    Types.DEVICE_OR_PORT_NAME: decoded_device_or_port_name,
    Types.CHANNEL_PREFIX: decoded_channel_prefix,
    Types.PORT_PREFIX: decoded_port_prefix,
    Types.TEMPO: decoded_tempo,
    Types.SMPTE_OFFSET: decoded_smpte_offset,
    Types.TIME_SIGNATURE: decoded_time_signature,
    Types.KEY_SIGNATURE: decoded_key_signature,
    Types.SEQUENCER_SPECIFIC_EVENT: decoded_sequencer_event,
}

assert Types.END_OF_TRACK not in MAPPING
assert len(MAPPING) == len(Types) - 1
assert all(x in {y.value for y in Types} for x in MAPPING)


def decoded(event):
    """ One of meta‑events. """
    # pylint: disable=too-many-branches
    if not isinstance(event, midi.MetaEvent):
        raise RuntimeError
    typ = event.type
    data = event.data

    if typ == Types.END_OF_TRACK:
        raise RuntimeError

    if typ in MAPPING:
        method = MAPPING[typ]
        result = method(data)
    else:
        result = UnknownMetaEvent(event=event)

    return result
