# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" MIDI initial definitions. """

from collections import namedtuple


# Event and realtime handlers
# ============================================================================

class Handler():
    """ Functional event and realtime handler.

    Use sie-effect on itself as a shortcut instead of requiring the caller
    to return a derived handler.

    """

    def __init__(self, acc_zero, ev_handler, rt_handler):
        self._acc = acc_zero
        self._ev_handler = ev_handler
        self._rt_handler = rt_handler

    @property
    def acc(self):
        """ The accumulator. """
        return self._acc

    def handle_ev(self, time, event):
        """ Apply `ev_handler` on `acc` then update `acc`. """
        self._acc = self._ev_handler(self._acc, time, event)

    def handle_rt(self, event):
        """ Apply `rt_handler` on `acc` then update `acc`. """
        self._acc = self._rt_handler(self._acc, event)


# Raw events
# ============================================================================

# There is no ChannelEvent, there is a ModeEvent and VoiceEvent instead.

MetaEvent = namedtuple("MetaEvent", ["type", "data"])
# May be further decoded by `meta_events.decoded`.

ModeEvent = namedtuple("ModeEvent", ["status", "data"])
# Distinguished channel event.
# May be further decoded by `mode_events.decoded`.
# There is a separate `VoiceEvent`.

RealtimeEvent = namedtuple("RealtimeEvent", ["status"])
# There is not further deconding.

SysExEvent = namedtuple("SysExEvent", ["data", "first", "last"])
# May be further decoded by `sysex_events.decoded`.

SystemEvent = namedtuple("SystemEvent", ["status", "data"])
# May be further decoded by `system_events.decoded`.
# There is a separate `RealtimeEvent`.
# Received regardless of unit's channel.

VoiceEvent = namedtuple("VoiceEvent", ["status", "data"])
# Distinguished channel event.
# May be further decoded by `voice_events.decoded`.
# Some may be received on the base channel only.
# There is a separate `ModeEvent`.


# Realtime
# ============================================================================

def is_realtime_status(status):
    """ If it's a realtime event status.

    Note in MIDI files, 0xFF is for meta‑events, and this function still
    returns `True` (realtime event) for 0xFF, which must be checked separately
    for MIDI files.

    """
    return status >= 0xF8 and status <= 0xFF


# Numbers
# ============================================================================


def hi7bits(number):
    """ Higher 7 bits of 14 bits. """
    return number // 128


def le14bits(data0, data1):
    """ 14 bits little‑endian from two bytes. """
    return data1 * 128 + data0


def be14bits(data0, data1):
    # FIXME: check uses.
    """ 14 bits big‑endian from two bytes. """
    return data0 * 128 + data1


def be16bits(data0, data1):
    """ 16 bits big‑endian from two bytes. """
    return data0 * 256 + data1


def le21bits(data0, data1, data2):
    """ 21 bits little‑endian from four bytes. """
    result = data2 * 128 + data1
    result = result * 128 + data0
    return result


def be24bits(data0, data1, data2):
    """ 24 bits big‑endian from three bytes. """
    return data0 * 65536 + data1 * 256 + data2


def le28bits(data0, data1, data2, data3):
    """ 28 bits little‑endian from four bytes. """
    result = data3 * 128 + data2
    result = result * 128 + data1
    result = result * 128 + data0
    return result
