# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" MIDI Channel Mode Events

A specialized control change channel event.

See also `mode_names`.

"""

from collections import namedtuple
from enum import (Enum, IntEnum)

from . import channel_events
from . import midi


# Constants
# ============================================================================

# Ranges
# ----------------------------------------------------------------------------

CONTROLLER_FIRST = channel_events.MODE_CONTROLLER_FIRST
CONTROLLER_LAST = 128
assert CONTROLLER_FIRST <= CONTROLLER_LAST
# `CONTROLLER_FIRST`..`CONTROLLER_LAST` makes‑up a range.


# Controllers
# ----------------------------------------------------------------------------

class Controllers(IntEnum):

    """ MIDI channel mode controllers. """

    ALL_SOUND_OFF = 120
    RESET_ALL_CONTROLLERS = 121
    LOCAL_CONTROL = 122
    ALL_NOTES_OFF = 123  # Should not override sustain controller
    OMNI_OFF = 124  # + all notes off
    OMNI_ON = 125  # + all notes off
    MONO_ON = 126  # Channels + all notes off
    POLY_ON = 127  # + mono off and all notes off


assert len({x.value for x in Controllers}) == len(Controllers)
# `Controllers` contains all unique items.

assert all(
    x.value >= CONTROLLER_FIRST and
    x.value <= CONTROLLER_LAST
    for x in Controllers)
# All `Controllers` items are in `CONTROLLER_FIRST`..`CONTROLLER_LAST`.


# Simple modes named after their controllers
# ----------------------------------------------------------------------------

class Modes(Enum):

    """ MIDI channel mode command zero based indexes. """

    ALL_SOUND_OFF = 0
    RESET_ALL_CONTROLLERS = 1
    LOCAL_CONTROL = 2
    ALL_NOTES_OFF = 3
    OMNI_OFF = 4
    OMNI_ON = 5
    # MONO_ON has its own type, it's not a simple mode.
    POLY_ON = 6


assert {x.name for x in Modes} == {x.name for x in Controllers} - {"MONO_ON"}
# Redefines all names from `Controllers` but `MONO_ON`.
assert len({x.value for x in Modes}) == len(Modes)
# `Modes` contains all unique items.
assert all(x.value >= 0 and x.value < len(Modes) for x in Modes)
# … and items form a zero based enumeration.


# Controllers to simple mode
# ----------------------------------------------------------------------------

MODES = {
    Controllers.ALL_SOUND_OFF: Modes.ALL_SOUND_OFF,
    Controllers.RESET_ALL_CONTROLLERS: Modes.RESET_ALL_CONTROLLERS,
    Controllers.LOCAL_CONTROL: Modes.LOCAL_CONTROL,
    Controllers.ALL_NOTES_OFF: Modes.ALL_NOTES_OFF,
    Controllers.OMNI_OFF: Modes.OMNI_OFF,
    Controllers.OMNI_ON: Modes.OMNI_ON,
    Controllers.POLY_ON: Modes.POLY_ON,
}

assert {x for x in MODES} == {x for x in Controllers} - {Controllers.MONO_ON}
assert {MODES[x] for x in MODES} == {x for x in Modes}


# Types
# ============================================================================

Mode = namedtuple("Mode", ["base_channel", "mode"])
Mono = namedtuple("Mono", ["base_channel", "channels"])


# Per Type
# ============================================================================

def decoded_mode(channel, controller):
    """ `Mode` for `controller`. """
    if channel >= channel_events.N_CHANNELS:
        raise ValueError
    if controller not in MODES:
        raise RuntimeError
    mode = MODES[controller]
    return Mode(base_channel=channel, mode=mode)


def decoded_mono(channel, channels):
    """ `Mono`. """
    if channel >= channel_events.N_CHANNELS:
        raise ValueError
    if channels > channel_events.N_CHANNELS:
        raise ValueError
    return Mono(base_channel=channel, channels=channels)


# Main
# ============================================================================

def decoded(event):
    """ One of channel mode events. """
    if not isinstance(event, midi.ModeEvent):
        raise RuntimeError
    data = event.data
    status = event.status
    channel = channel_events.channel(status)
    controller = data[0]
    # Data[1] is ignored except for `Mono`. When ignored, it should be
    # zero, but this is not checked so no error is raised when it is not.
    if controller == Controllers.MONO_ON:
        result = decoded_mono(channel=channel, channels=data[1])
    else:
        result = decoded_mode(channel=channel, controller=controller)
    return result
