# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" MIDI Realtime Events. """


from collections import namedtuple
from enum import (Enum, IntEnum)

from . import midi


# Constants
# ============================================================================

FIRST_STATUS = 0xF8
LAST_STATUS = 0xFF


class Statuses(IntEnum):

    """ Realtime event statuses. """

    CLOCK = 0xF8
    TICK = 0xF9
    START = 0xFA
    CONTINUE = 0xFB
    STOP = 0xFC
    ACTIVE_SENSE = 0xFE  # Triggers timeout mechanism.
    RESET = 0xFF  # Not in MIDI files, where it's meta‑event.

    # 0xFD is undefined.


assert all(
    x.value >= FIRST_STATUS and
    x.value <= LAST_STATUS
    for x in Statuses)

assert len({x.value for x in Statuses}) == len(Statuses)


class Events(Enum):

    """ Realtime event indexes. """

    CLOCK = 0
    TICK = 1
    START = 2
    CONTINUE = 3
    STOP = 4
    ACTIVE_SENSE = 5
    RESET = 6


assert len(Events) == len(Statuses)
assert all(x.name in {y.name for y in Statuses} for x in Events)
assert len({x.value for x in Events}) == len(Events)
assert all(x.value >= 0 and x.value < len(Events) for x in Events)


# Type
# ============================================================================

Event = namedtuple("Event", ["event"])


# Main
# ============================================================================

MAPPING = {
    Statuses.CLOCK: Events.CLOCK,
    Statuses.TICK: Events.TICK,
    Statuses.START: Events.START,
    Statuses.CONTINUE: Events.CONTINUE,
    Statuses.STOP: Events.STOP,
    Statuses.ACTIVE_SENSE: Events.ACTIVE_SENSE,
    Statuses.RESET: Events.RESET,
}

assert len(MAPPING) == len(Statuses)
assert all(x.value in MAPPING for x in Statuses)
assert all(MAPPING[x].value in {y.value for y in Events} for x in MAPPING)
assert len({MAPPING[x] for x in MAPPING}) == len(Events)


def decoded(event):
    """ Decoded realtime event. """
    if not isinstance(event, midi.RealtimeEvent):
        raise RuntimeError
    status = event.status
    if status not in MAPPING:
        raise RuntimeError
    event_index = MAPPING[status]
    return Event(event=event_index)
