# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" Standard MIDI file reader. """

from collections import namedtuple

import smpte

from . import midi
from . import channel_events
from . import system_events


# Input
# ============================================================================

# Streams
# ----------------------------------------------------------------------------

class Stream():
    """ Functional stream.

    Use sie-effect on itself as a shortcut instead of requiring the caller
    to return a derived stream.

    """

    def __init__(self, stream):
        self._stream = stream
        self._remaining = 0

    @property
    def remaining(self):
        """ Remaining. """
        return self._remaining

    def set_max_size(self, max_size):
        """ Set max size. """
        self._remaining = max_size

    def bits(self):
        """ Return current byte updating `max_size`. """
        if self._remaining == 0:
            raise ValueError
        data = self._stream.read(1)
        if len(data) != 1:
            raise ValueError
        self._remaining -= 1
        return data[0]


# SMF and MIDI Bytes
# ----------------------------------------------------------------------------

def read_smf_byte(stream):
    """ Read one SMF specific byte. """
    return stream.bits()


def read_midi_byte(stream, handler):
    """ Read one MIDI specific byte.

    May trigger a realtime event.

    """

    while True:
        bits = stream.bits()
        if not midi.is_realtime_status(bits):
            break
        if bits == 0xFF:  # Only in MIDI files.
            break
        event = midi.RealtimeEvent(status=bits)
        handler.handle_rt(event)

    return bits


# Data Blocks
# ----------------------------------------------------------------------------

DATA_MODE_7BITS = 0
DATA_MODE_SYSEX = 1
DATA_MODE_8BITS = 2


def read_data(buffer, stream, data_size, mode, handler):
    """ Read data, validating for requested mode. """
    if mode != DATA_MODE_8BITS and handler is None:
        raise RuntimeError
    while len(buffer) < data_size:
        if mode == DATA_MODE_8BITS:
            bits = read_smf_byte(stream)
        else:
            bits = read_midi_byte(stream=stream, handler=handler)
            if bits >= 0x80:
                if mode == DATA_MODE_7BITS:
                    raise ValueError
                elif mode == DATA_MODE_SYSEX:
                    if len(buffer) + 1 != data_size:
                        raise ValueError
        buffer.append(bits)
    return bytes(buffer)


# Headers
# ============================================================================

def read_four_smf_bytes(stream):
    """ Read four SMF specific bytes. """
    return read_data(bytearray(), stream, 4, DATA_MODE_8BITS, None)


def read_smf_int(stream, size):
    """ Read big‑endian integer of a fixed size in SMF. """
    result = 0
    count = 0
    while count < size:
        # BE integer.
        result *= 256
        result += read_smf_byte(stream)
        count += 1
    return result


ChunkHeader = namedtuple("ChunkHeader", ["id", "length"])


def read_chunk_header(stream):
    """ Read ChunkHeader. """
    return ChunkHeader(
        id=read_four_smf_bytes(stream),
        length=read_smf_int(stream, 4))


MTHDChunk = namedtuple("MTHDChunk", ["format", "num_tracks", "division"])
MTHD_ID = b"MThd"
MTHD_LENGTH = 6
MTRK_ID = b"MTrk"


PPQNDivision = namedtuple("PPQNDivision", ["value"])
SMPTEDivision = namedtuple("SMPTEDivision", ["rate", "subframes"])


def time_division_from_bytes(data):
    """ `SMPTEDivision` or `PPQNDivision` from bytes. """
    if len(data) != 2:
        raise RuntimeError
    if data[0] < 0x80:
        # BE integer from SMF: MSB then LSB.
        value = data[0] * 256 + data[1]
        result = PPQNDivision(value=value)
    else:
        rate_code = data[0]
        subframes = data[1]
        if rate_code == 256 - 24:
            rate = smpte.RATE_24FPS
        elif rate_code == 256 - 25:
            rate = smpte.RATE_25FPS
        elif rate_code == 256 - 29:
            rate = smpte.RATE_30FPS_DF
        elif rate_code == 256 - 30:
            rate = smpte.RATE_30FPS
        else:
            raise ValueError
        result = SMPTEDivision(rate=rate, subframes=subframes)
    return result


def read_mthd_chunk(stream):
    """ Read `MTHDChunk`. """
    stream = Stream(stream)
    stream.set_max_size(14)
    chunk_header = read_chunk_header(stream)
    if chunk_header.id != MTHD_ID:
        raise ValueError
    if chunk_header.length != MTHD_LENGTH:
        raise ValueError
    fmt = read_smf_int(stream, 2)
    if fmt > 2:
        raise ValueError
    num_tracks = read_smf_int(stream, 2)
    time_division_data = read_data(
        buffer=bytearray(),
        stream=stream,
        data_size=2,
        mode=DATA_MODE_8BITS,
        handler=None)
    division = time_division_from_bytes(time_division_data)
    return MTHDChunk(format=fmt, num_tracks=num_tracks, division=division)


# VarLen Data
# ============================================================================

def read_var_len(stream):
    """ Read SMF variable length. """

    count = 0
    result = 0

    while True:

        if count == 4:
            raise ValueError

        bits = read_smf_byte(stream)
        count += 1

        if bits >= 0x80:
            bits -= 0x80
            last = False
        else:
            last = True

        # BE integer
        result *= 128
        result += bits
        if last:
            break

    return result


def read_var_len_data(stream, mode, handler):
    """ Read data of variable length. """
    data_size = read_var_len(stream)
    buffer = bytearray()
    return read_data(
        buffer=buffer,
        stream=stream,
        data_size=data_size,
        mode=mode,
        handler=handler)


# MIDI and SMF Events
# ============================================================================

def channel_event(status, data):
    """ Voice or Mode Event, depending on status and data. """
    if channel_events.is_mode_event(status, data):
        result = midi.ModeEvent(status=status, data=data)
    else:
        result = midi.VoiceEvent(status=status, data=data)
    return result


def skip_data_bytes(stream, handler):
    """ Skip data bytes. Used when data without a running status. """
    while True:
        bits = read_midi_byte(stream, handler)
        if bits >= 0x80:
            result = bits
            break
    return result


def read_running_status_event(
        status,
        first_bits,
        stream,
        handler):
    """ Read channel event using running status.

    Invoked from `read_event` only.

    """
    data_size = channel_events.data_size(status)
    buffer = bytearray()
    buffer.append(first_bits)
    data = read_data(
        buffer=buffer,
        stream=stream,
        data_size=data_size,  # Note there is already one byte in buffer.
        mode=DATA_MODE_7BITS,
        handler=handler)
    return channel_event(status=status, data=data)


def read_channel_event(status, stream, handler):
    """ Read channel event with explicit status.

    Invoked from `read_event` only.

    """
    data_size = channel_events.data_size(status)
    buffer = bytearray()
    data = read_data(
        buffer=buffer,
        stream=stream,
        data_size=data_size,
        mode=DATA_MODE_7BITS,
        handler=handler)
    return channel_event(status=status, data=data)


def read_system_event(status, stream, handler):
    """ Read system event.

    Invoked from `read_event` only.

    """
    data_size = system_events.data_size(status)
    buffer = bytearray()
    data = read_data(
        buffer=buffer,
        stream=stream,
        data_size=data_size,
        mode=DATA_MODE_7BITS,
        handler=handler)
    return midi.SystemEvent(status=status, data=data)


def read_sysex_packet_event(sysex_cont, status, stream, handler):
    """ Read SysEx packet event.

    Invoked from `read_event` only.

    """
    # On a MIDI connection, a SysEx may be terminated by any status byte
    # but the realtime ones.
    if status == 0xF7 and not sysex_cont:
        raise NotImplementedError
    data = read_var_len_data(
        stream=stream,
        mode=DATA_MODE_SYSEX,
        handler=handler)
    first = status == 0xF0
    if data:  # len(data) != 0
        if data[-1] == 0xF7:
            last = True
            data = data[:-1]
        else:
            last = False
    return midi.SysExEvent(data=data, first=first, last=last)


def read_meta_event(stream):
    """ Read meta‑event.

    Invoked from `read_event` only.

    """
    meta_event_type = read_smf_byte(stream)
    if meta_event_type >= 128:
        raise ValueError
    data = read_var_len_data(
        stream=stream,
        mode=DATA_MODE_8BITS,
        handler=None)
    return midi.MetaEvent(type=meta_event_type, data=data)


def read_event(running_status, sysex_cont, stream, handler):
    """ Read SMF or MIDI event. """
    first_bits = read_midi_byte(stream, handler)

    if sysex_cont and first_bits != 0xF7:
        raise ValueError

    if first_bits >= 0x80:  # New status
        status = first_bits
    elif running_status is not None:  # Data of running status
        status = running_status
    else:  # Skip data until new status.
        first_bits = skip_data_bytes(stream, handler)
        status = first_bits

    if first_bits < 0x80:
        result = read_running_status_event(
            status=status,
            first_bits=first_bits,
            stream=stream,
            handler=handler)
    elif status < 0xF0:
        result = read_channel_event(
            status=status,
            stream=stream,
            handler=handler)
    elif status == 0xF0 or status == 0xF7:  # Escape not implemented.
        result = read_sysex_packet_event(
            sysex_cont=sysex_cont,
            status=status,
            stream=stream,
            handler=handler)
    elif status == 0xFF:
        result = read_meta_event(stream)
    else:
        result = read_system_event(
            status=status,
            stream=stream,
            handler=handler)

    return result


# Track Chunk
# ============================================================================

def skip_chunk(stream, chunk_cheader):
    """ Skip chunk (used for non‑track chunk). """
    count = chunk_cheader.length
    while count > 0:
        read_smf_byte(stream)
        count -= 1


def read_mtrk_chunk(stream, handler):
    """ Read events. """

    stream = Stream(stream)
    stream.set_max_size(8)

    chunk_header = read_chunk_header(stream)
    if chunk_header.id != MTRK_ID:
        skip_chunk(stream, chunk_header)
        return False

    time = 0
    running_status = None
    sysex_cont = False

    stream.set_max_size(chunk_header.length)

    while True:
        delta_time = read_var_len(stream)
        time += delta_time
        event = read_event(
            running_status=running_status,
            sysex_cont=sysex_cont,
            stream=stream,
            handler=handler)
        if isinstance(event, midi.VoiceEvent):
            running_status = event.status
        elif isinstance(event, midi.ModeEvent):
            running_status = event.status
        elif isinstance(event, midi.MetaEvent):
            running_status = None
            if event.type == 0x2F:  # pylint: disable=no-member
                if event.data:  # len(event.data) != 0
                    raise ValueError
                break
        elif isinstance(event, midi.SysExEvent):
            running_status = None
            sysex_cont = not event.last  # pylint: disable=no-member
        else:
            running_status = None

        handler.handle_ev(time, event)

    if stream.remaining != 0:
        raise ValueError
    return True
