# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" MIDI State. """

from enum import Enum

from . import controller_events
from . import midi
from . import mode_events
from . import voice_events

from .controller_events import (
    N_COMMANDS,
    N_SELECTORS,
    N_SLIDERS_7_BITS,
    N_SLIDERS_14_BITS,
    N_SWITCHES)

N_CHANNELS = 16

GM1_MINIMUM_VOICES = 24
GM1_PERCUSSION_CHANNEL = 9

DEFAULT_BASE_SYSEX_CHANNEL = 0
DEFAULT_BASE_VOICE_CHANNEL = 0
DEFAULT_CHANNEL_PRESSURE = 63
DEFAULT_ENABLED = True
DEFAULT_IGNORE_PROGRAM_CHANGE = False
DEFAULT_KEY_PRESSURE = 63
DEFAULT_LOCAL_CONTROL = True
DEFAULT_MIDI_THROUGH = False
DEFAULT_MULTIMODE = True
DEFAULT_NBO_MONO_CHANNELS = 16
DEFAULT_NBO_VOICES = 16  # There may be static allocation per channel.
DEFAULT_OMNI = True  # And `DEFAULT_POLY` is `True` (hence MIDI Mode 1)
DEFAULT_PITCH_BEND = 0
DEFAULT_POLY = True  # And `DEFAULT_OMNI` is `True` (hence MIDI Mode 1)
DEFAULT_PROGRAM = 0
DEFAULT_SELECTOR_VALUE = 0
DEFAULT_SLIDER_VALUE = 0
DEFAULT_VELOCITY = voice_events.DEFAULT_VELOCITY  # 64, Mezzo‑forte


def warning(_message):
    """ On event warning (when an event was unexpected). """
    pass


class ADSR(Enum):

    """ ADSR phase constants. """

    Attack = 1
    Decay = 2
    Sustain = 3
    Release = 4


class Note():

    """ MIDI note playing. """

    @staticmethod
    def transition(note):
        """ Phase transition from `note` which may be `None`. """
        if note is None:
            result = ADSR.Attack
        elif note.phase == ADSR.Attack:
            result = ADSR.Sustain
        elif note.phase == ADSR.Decay:
            result = ADSR.Sustain
        elif note.phase == ADSR.Sustain:
            result = ADSR.Sustain
        elif note.phase == ADSR.Release:
            result = ADSR.Attack
        return result

    def __init__(self, key, velocity, transition_from=None):
        """ Bar """
        self._key = key
        self._phase = self.transition(transition_from)
        self._pressure = DEFAULT_KEY_PRESSURE
        # May trigger effect, shift timbre, etc.
        self._velocity = velocity

    @property
    def key(self):
        """ Key (whatever the scale is). """
        return self._key

    @property
    def phase(self):
        """ ADSR phase. """
        return self._phase

    @property
    def velocity(self):
        """ Velocity. """
        return self._velocity

    def restart(self, velocity):
        """ Get back to `ADSR.Attack` phase with `velocity`. """
        self._phase = ADSR.Attack
        self._velocity = velocity

    def stop(self, velocity):
        """ Note goes to `ADSR.Release` phase. """
        self._phase = ADSR.Release
        self._velocity = velocity

    def set_pressure(self, pressure):
        """ Set note pressure if it differs from the old one. """
        if pressure != self._pressure:
            self._pressure = pressure


class Control():
    """ Local control of a MIDI unit. """
    # May separate group of keys by channel.
    pass


class MonoChannel():

    """ Mono‑channel in Omni Off, internal to a multi‑timbral device  """

    def __init__(self, base_channel):
        self._base_channel = base_channel
        self._coarse_sliders = [DEFAULT_SLIDER_VALUE] * N_SLIDERS_7_BITS
        self._commands = [None] * N_COMMANDS
        self._enabled = DEFAULT_ENABLED
        self._fine_sliders = [DEFAULT_SLIDER_VALUE] * N_SLIDERS_14_BITS
        self._ignore_program_change = DEFAULT_IGNORE_PROGRAM_CHANGE
        self._local_control = DEFAULT_LOCAL_CONTROL
        self._multi_mode = DEFAULT_MULTIMODE
        self._nbo_mono_channels = DEFAULT_NBO_MONO_CHANNELS  # See note.
        self._nbo_voices = DEFAULT_NBO_VOICES
        self._notes_playing = dict()
        self._omni = DEFAULT_OMNI
        self._pitch_bend = DEFAULT_PITCH_BEND
        self._poly = DEFAULT_POLY
        self._pressure = DEFAULT_CHANNEL_PRESSURE
        self._program = DEFAULT_PROGRAM
        self._selectors = [DEFAULT_SELECTOR_VALUE] * N_SELECTORS
        self._switches = [None] * N_SWITCHES

        # Note: the `nbo_mono_channels` property gives a computed effective
        # value, `_nbo_mono_channels` must not be read directly, except in
        # the mentionned property.

    @property
    def base_channel(self):
        """ Channel number, zero based. """
        return self._base_channel

    def coarse_slider(self, index):
        """ Coarse slider value by index. """
        assert index < N_SLIDERS_7_BITS
        return self._coarse_sliders[index]

    @property
    def enabled(self):
        """ If responds to messages. Not part of MIDI. """
        return self._enabled

    def set_enabled(self, enabled):
        """ Set `enabled`. Not part of MIDI. """
        self._enabled = enabled

    def fine_slider(self, index):
        """ Fine slider value by index. """
        assert index < N_SLIDERS_14_BITS
        return self._fine_sliders[index]

    @property
    def ignore_program_change(self):
        """ If program changes are ignored. """
        return self._ignore_program_change

    def set_ignore_program_change(self, ignore):
        """ Set `ignore_program_change`. """
        self._ignore_program_change = ignore

    def is_global_channel(self, channel):
        """ If the channel correspond to a global channel. """
        # FIXME: for controller messages only or voice messages too?
        return (channel + 1) % 16 == self._base_channel

    @property
    def local_control(self):
        """ If local control is On. """
        return self._local_control

    @property
    def mode(self):
        """ MIDI Mode 1-4. """
        result = 1 if self._omni else 3
        result += 0 if self._poly else 1
        return result

    @property
    def mono(self):
        """ If in Mono mode. """
        return not self._poly

    @property
    def multi_mode(self):
        """ If in Multi Mode. """
        return self._multi_mode

    @property
    def nbo_channels(self):
        """ Number of channels. """
        return 1

    @property
    def nbo_mono_channels(self):
        """ Numer of channels for MIDI Mode 4. """
        max_channels = 16 - self._base_channel  # Base channel is zero based.
        max_channels = min(max_channels, self._nbo_voices)
        max_channels = min(max_channels, self.nbo_channels)
        if self._nbo_mono_channels == 0:
            result = max_channels
        else:
            result = min(max_channels, self._nbo_mono_channels)
        return result

    @property
    def nbo_voices(self):
        """ Numer of voices. """
        # FIXME: see it as static allocation per channel?
        # If change is allowed, than must care about Mono Channels.
        return self._nbo_voices

    @property
    def omni(self):
        """ If Omni On. """
        # When On, All Notes Off should be ignored. The same for Reset All
        # Controllers.
        return self._omni

    @property
    def poly(self):
        """ If in Poly mode. """
        # When On and Omni Off, All Notes Off should apply to the base
        # channel only. The same for Reset All Controllers.
        # When Off and Omni Off, All Notes Off should apply to the channel
        # from which the message was received. The same for Reset All
        # Controllers.
        return self._poly

    def selector(self, index):
        """ Selector value by index. """
        assert index < N_SELECTORS
        return self._selectors[index]

    def switch(self, index):
        """ Switch value by index. """
        assert index < N_SWITCHES
        return self._switches[index]

    @staticmethod
    def updated_14bits_entry(items, index, nibble, msb):
        """ Update an enty of a 14 bits number list. """
        if msb:
            hi = nibble
            lo = 0
        else:
            hi = midi.hi7bits(items[index])
            lo = nibble
        value = midi.be14bits(hi, lo)
        items[index] = value

    def transmit(self, event):
        """ Send MIDI event. """
        # On an unhandled MIDI Mode, message should be ignored.
        # MIDI Mode and base channel may be different for transmitting.
        # FIXME: automatic Note Off with transmitter in Mono mode?
        pass

    def _handle_control_change(self, raw_event):
        """ Handle `channel_events.ControlChange`. """
        assert isinstance(raw_event, voice_events.ControlChange)
        event = controller_events.decoded(raw_event)
        if isinstance(event, controller_events.Slider7Bits):
            index = event.index
            value = event.value
            self._coarse_sliders[index] = value
        elif isinstance(event, controller_events.Command):
            pass
        elif isinstance(event, controller_events.Slider14Bits):
            self.updated_14bits_entry(
                items=self._fine_sliders,
                index=event.index,
                nibble=event.value,
                msb=event.msb)
        elif isinstance(event, mode_events.Mono):
            nbo_mono_channels = event.channels
            if nbo_mono_channels != self._nbo_mono_channels:
                self._nbo_mono_channels = nbo_mono_channels
        elif isinstance(event, controller_events.Selector):
            self.updated_14bits_entry(
                items=self._selectors,
                index=event.index,
                nibble=event.value,
                msb=event.msb)
        elif isinstance(event, controller_events.Switch):
            index = event.index
            switch_on = event.on
            self._switches[index] = switch_on
        else:
            raise RuntimeError

    def _handle_note_off(self, event):
        """ Note Off. """
        assert isinstance(event, voice_events.Note)
        key = event.key
        velocity = event.velocity
        if key in self._notes_playing:
            note = self._notes_playing[key]
            note.stop(velocity)
        else:
            warning("Stopping a note which was not playing")

    def _handle_note_on_poly(self, event):
        """ Note On when Poly: Poly Voice Assigner. """
        assert isinstance(event, voice_events.Note)
        assert self._poly is True
        key = event.key
        velocity = event.velocity
        if key not in self._notes_playing:
            note = Note(key=key, velocity=velocity)
            self._notes_playing[key] = note
        else:
            note = self._notes_playing[key]
            note.restart(velocity)

    def _handle_note_on_mono_onmi_on(self, event):
        """ Note On when Mono: Mono Voice Assigner when Omni On. """
        assert isinstance(event, voice_events.Note)
        assert self._poly is False
        key = event.key
        velocity = event.velocity
        transition_from = None
        if self._notes_playing:  # len(self._notes_playing) != 0
            playing_key = self._notes_playing.keys()[0]
            transition_from = self._notes_playing[playing_key]
            del self._notes_playing[playing_key]
        note = Note(
            key=key,
            velocity=velocity,
            transition_from=transition_from)
        self._notes_playing[key] = note

    def _handle_note_on_mono_onmi_off(self, event):
        """ Note On when Mono: Mono Voice Assigner when Omni Off. """
        assert isinstance(event, voice_events.Note)
        assert self._poly is False

    def _handle_note_on_mono(self, event):
        """ Note On when Mono: Mono Voice Assigner. """
        assert isinstance(event, voice_events.Note)
        assert self._poly is False
        assert len(self._notes_playing) <= 1

    def _handle_note(self, event, internal=False):
        """ Handle `channel_events.Note`. """
        # FIXME: internal is also about All Notes Off, which must not
        # apply to internal notes.
        assert isinstance(event, voice_events.Note)
        if not (internal and self._local_control is False):
            if event.on is False:
                self._handle_note_off(event)
            elif self._poly:
                self._handle_note_on_poly(event)
            else:
                self._handle_note_on_mono(event)
        # FIXME: transmit to MIDI Out if internal, and to
        # MIDI Through if not internal.

    def _handle_note_pressure(self, event):
        """ Handle `channel_events.KeyPressure`. """
        assert isinstance(event, voice_events.KeyPressure)
        key = event.key
        pressure = event.value
        if key in self._notes_playing:
            note = self._notes_playing[key]
            note.set_pressure(pressure)
        else:
            warning("Pressure on a note which is not playing")

    def handle_voice_event(self, raw_event):
        """ Handle `midi.VoiceEvent`. """
        # Receiver should honnor a status event if the previous one was
        # not completed/handled. Receiver should ignore unhandled statuses.
        # On an unhandled MIDI Mode, message should be ignored.
        assert isinstance(raw_event, midi.VoiceEvent)
        if self._enabled:
            if self._omni or self._base_channel == raw_event.channel:
                # Mono channel unit, therefore the comparison to base channel.
                event = channel_events.decoded(raw_event)
                if isinstance(event, voice_events.ChannelPressure):
                    self._pressure = event.value
                elif isinstance(event, voice_events.ControlChange):
                    self._handle_control_change(event)
                elif isinstance(event, voice_events.KeyPressure):
                    self._handle_note_pressure(event)
                elif isinstance(event, voice_events.Note):
                    self._handle_note(event)
                elif isinstance(event, voice_events.PitchBend):
                    self._pitch_bend = event.value
                elif isinstance(event, voice_events.ProgramChange):
                    self._program = event.program
                else:
                    raise RuntimeError

    def handle_mode_event(self, raw_event):
        """ Handle `midi.ModeEvent`. """
        pass


class Device():

    """ MIDI Device state. """

    def __init__(self, name):
        self._base_sysex_channel = DEFAULT_BASE_SYSEX_CHANNEL
        self._base_voice_channel = DEFAULT_BASE_VOICE_CHANNEL
        self._channels = [MonoChannel(i) for i in range(0, N_CHANNELS)]
        self._midi_through = DEFAULT_MIDI_THROUGH
        self._name = name

    @property
    def omni(self, index):
        """ If Omni On. """
        return self._omni

    def _handle_midi_mode(raw_event):
        assert isinstance(raw_event, midi.ModeEvent)
        event = mode_events.decoded(event)
        if isinstance(event, mode_events.Mode):
            index = event.index
            if index == mode_events.Modes.OMNI_OFF:
                pass
            elif index == mode_events.Modes.OMNI_ON:
                pass
            elif index == mode_events.Modes.POLY_ON:
                pass
            else:
                raise RuntimeError
        elif isinstance(event, mode_events.Mono):
            pass
        else:
            raise RuntimeError

    def handle_event(self, event):
        """ Handle event from `midi`. """
        if isinstance(event, midi.MetaEvent):
            pass
        elif isinstance(event, midi.ModeEvent):
            pass
        elif isinstance(event, midi.RealtimeEvent):
            pass
        elif isinstance(event, midi.SysExEvent):
            pass
        elif isinstance(event, midi.SystemEvent):
            pass
        elif isinstance(event, midi.VoiceEvent):
            pass
        else:
            raise RuntimeError
