# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" SysEx Events. """

from collections import namedtuple

from . import midi
from . import manufacturers


# Constant
# ============================================================================

BROADCAST_CHANNEL = 127
# For `Universal.channel`. Checked by `is_broadcast`.


def is_broadcast(event):
    """ If the channel indicates broadcast. """
    if not isinstance(event, Universal):
        raise RuntimeError
    return event.channel == BROADCAST_CHANNEL


# Types
# ============================================================================

Head = namedtuple("Head", ["manufacturer", "data", "last"])
# `manufacturer` is an index, not the manufacturer ID. See `manufacturers`.

Tail = namedtuple("Tail", ["data", "last"])

Universal = namedtuple(
    "Universal",
    ["realtime", "channel", "sub_id", "data", "last"])
# May be further decoded with `universal_sysex_events.decoded`.
# This event may be followed by a `Tail`, just like with `Head`.


# Main
# ============================================================================

def decoded(event):
    """ `SysExHead`/`Universal` or `SysExTail`. """
    if not isinstance(event, midi.SysExEvent):
        raise RuntimeError
    if not event.first:
        data = event.data
        last = event.last
        result = Tail(data=data, last=last)
    elif event.data[0] in [0x7E, 0x7F]:  # Universal SysEx
        if len(event.data) < 4:
            raise ValueError
        last = event.last
        realtime = event.data[0] == 0x7F
        channel = event.data[1]
        sub_id = event.data[2]
        data = event.data[3:]
        result = Universal(
            realtime=realtime,
            channel=channel,
            sub_id=sub_id,
            data=data,
            last=last)
    else:
        manufacturer_data = manufacturers.decoded_data(event.data)
        manufacturer = manufacturer_data.manufacturer
        data = manufacturer_data.data
        last = event.last
        result = Head(
            manufacturer=manufacturer,
            data=data,
            last=last)
    return result
