# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" MIDI System Events. """

from collections import namedtuple
from enum import IntEnum

from . import midi


# Constants
# ============================================================================

FIRST_STATUS = 0xF1
LAST_STATUS = 0xF6


class Statuses(IntEnum):

    """ System event constants. """

    QUARTER_FRAME = 0xF1  # Part of MTC
    SONG_POSITION_POINTER = 0xF2  # What to play on `START`.
    SONG_SELECT = 0xF3  # What to play on `START`.
    TUNE_REQUEST = 0xF6

    # 0xF4 and 0xF5 are undefined.


assert all(
    x.value >= FIRST_STATUS and
    x.value <= LAST_STATUS
    for x in Statuses)

assert len({x.value for x in Statuses}) == len(Statuses)


# Types
# ============================================================================

QuarterFrame = namedtuple("QuarterFrame", ["tag", "nibble"])
SongPosition = namedtuple("SongPosition", ["value"])
SongSelect = namedtuple("SongSelect", ["index"])
TuneRequest = namedtuple("TuneRequest", [])


def data_size(status):
    """ Data size for MIDI status not associated to a channel. """
    if status == Statuses.QUARTER_FRAME:
        result = 1
    elif status == Statuses.SONG_POSITION_POINTER:
        result = 2
    elif status == Statuses.SONG_SELECT:
        result = 1
    elif status == Statuses.TUNE_REQUEST:
        result = 0
    else:  # 0xF4 and 0xF5 are undefined.
        raise ValueError
    return result


# Per Type
# ============================================================================

def decoded_quarter_frame(data):
    """ `QuarterFrame` from data. """
    if len(data) != 1:
        raise RuntimeError
    tag = data[0] // 16
    nibble = data[0] % 16
    if tag >= 8:
        raise ValueError
    return QuarterFrame(tag=tag, nibble=nibble)


def decoded_song_position(data):
    """ Song position from data. """
    if len(data) != 2:
        raise RuntimeError
    value = midi.le14bits(data[0], data[1])
    return SongPosition(value=value)


def decoded_song_select(data):
    """ Song select from data. """
    if len(data) != 1:
        raise RuntimeError
    index = data[0]
    return SongSelect(index=index)


def decoded_tune_request(data):
    """ Tune request from data. """
    if data:  # len(data) != 0
        raise RuntimeError
    return TuneRequest()


# Main
# ============================================================================

def decoded(event):
    """ One of system events. """
    if not isinstance(event, midi.SystemEvent):
        raise RuntimeError
    status = event.status
    data = event.data
    if status == Statuses.QUARTER_FRAME:
        result = decoded_quarter_frame(data)
    elif status == Statuses.SONG_POSITION_POINTER:
        result = decoded_song_position(data)
    elif status == Statuses.SONG_SELECT:
        result = decoded_song_select(data)
    elif status == Statuses.TUNE_REQUEST:
        result = decoded_tune_request(data)
    else:  # 0xF0 and 0xF7 are not handled here.
        raise RuntimeError

    return result
