# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" Universal SysEx Events. """

from collections import namedtuple
from enum import (Enum, IntEnum)

from . import manufacturers
from . import midi
from . import sysex_events


# Events
# ============================================================================

# Non‑Realtime
# ----------------------------------------------------------------------------

DumpHeader = namedtuple(
    "DumpHeader",
    ["channel",
     "wave_number",
     "resolution",
     "period",
     "length",
     "start",
     "end",
     "kind"])

DumpLoopPointsRequest = namedtuple(
    "DumpLoopRequest",
    ["channel",
     "wave_number",
     "loop_number"])

DumpLoopPointsTransmit = namedtuple(
    "DumpLoopPointsTransmit",
    ["channel",
     "loop_points"])

DumpPacket = namedtuple(
    "DumpPacket",
    ["channel",
     "packet_number",
     "data",
     "error"])

DumpRequest = namedtuple(
    "DumpRequest",
    ["channel",
     "wave_number"])

GMMode = namedtuple(
    "GMMode",
    ["channel",
     "level",
     "on"])

Handshake = namedtuple(
    "Handshake",
    ["channel",
     "packet_number",
     "status"])

IdentityReply = namedtuple(
    "IdentityReply",
    ["channel",
     "manufacturer",
     "familly",
     "model",
     "version"])

IdentityRequest = namedtuple(
    "IdentityRequest",
    ["channel"])

LoopPoint = namedtuple(
    "LoopPoint",
    ["wave_number",
     "loop_number",
     "kind",
     "start",
     "end"])


# Realtime
# ----------------------------------------------------------------------------

MasterVolume = namedtuple("MasterVolume", ["channel", "volume"])


# Sub ID#1 and Sub ID#2
# ============================================================================

class ID1s(IntEnum):

    """ Sub ID/Sub ID#1 of Universal SysEx. """

    SAMPLE_DUMP_HEADER = 0x01
    SAMPLE_DUMP_PACKET = 0x02
    SAMPLE_DUMP_REQUEST = 0x03
    DEVICE_CONTROL = 0x04  # Uses Sub ID#2
    SAMPLE_LOOP_POINTS = 0x05  # Uses Sub ID#2
    GENERAL_INFORMATION = 0x06  # Uses Sub ID#2
    GM_MODE = 0x09
    HANDSHAKE_WAIT = 0x7C
    HANDSHAKE_CANCEL = 0x7D
    HANDSHAKE_NAK = 0x7E
    HANDSHAKE_ACK = 0x7F


class ID2s(IntEnum):

    """ Sub ID#2 of Universal SysEx. """

    # With `ID1s.DEVICE_CONTROL`
    MASTER_VOLUME = 0x01

    # With `ID1s.SAMPLE_LOOP_POINTS`
    SAMPLE_LOOP_POINTS_TRANSMIT = 0x01
    SAMPLE_LOOP_POINTS_REQUEST = 0x02

    # With `ID1s.GENERAL_INFORMATION`
    IDENTITY_REQUEST = 0x01
    IDENTITY_REPLY = 0x02


# Event type check
# ============================================================================

def is_of_expected_type(event, realtime, sub_id, sub_id2=None):
    """ Check event type for assertion. """
    return (
        isinstance(event, sysex_events.Universal) and
        event.realtime == realtime and
        event.sub_id == sub_id and
        len(event.data) >= 1 and
        (sub_id2 is None or event.data[0] == sub_id2))


# Non‑realtime
# ============================================================================

# Constants
# ----------------------------------------------------------------------------

class Handshakes(Enum):

    """ Handshake statuses. """

    NAK = 0
    ACK = 1
    CANCEL = 2
    WAIT = 3


# Per Sub ID#1
# ----------------------------------------------------------------------------

def dump_header(event):
    """ Decoded Dump Header non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.SAMPLE_DUMP_HEADER):
        raise RuntimeError
    channel = event.channel
    data = event.data
    if len(data) != 16:
        raise ValueError
    wave_number = midi.le14bits(data[0], data[1])
    resolution = data[2]
    period = midi.le21bits(data[3], data[4], data[5])
    length = midi.le21bits(data[6], data[7], data[8])
    start = midi.le21bits(data[9], data[10], data[11])
    end = midi.le21bits(data[12], data[13], data[14])
    kind = data[15]
    return DumpHeader(
        channel=channel,
        wave_number=wave_number,
        resolution=resolution,
        period=period,
        length=length,
        start=start,
        end=end,
        kind=kind)


def dump_loop_points_request(event):
    """ Decoded Dump Loop Points Request non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.SAMPLE_LOOP_POINTS,
            sub_id2=ID2s.SAMPLE_LOOP_POINTS_REQUEST):
        raise RuntimeError
    channel = event.channel
    data = event.data[1:]
    if len(data) == 4:
        raise ValueError
    wave_number = midi.le14bits(data[0], data[1])
    loop_number = midi.le14bits(data[2], data[3])
    return DumpLoopPointsRequest(
        channel=channel,
        wave_number=wave_number,
        loop_number=loop_number)


def dump_loop_points_transmit(event):
    """ Decoded Dump Loop Points Transmit non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.SAMPLE_LOOP_POINTS,
            sub_id2=ID2s.SAMPLE_LOOP_POINTS_TRANSMIT):
        raise RuntimeError
    channel = event.channel
    data = event.data[1:]
    if not data:  # len(data) == 0
        raise ValueError
    if (len(data) % 11) != 0:
        raise ValueError
    count = len(data) // 11
    loop_points = [None] * count
    i = 0
    j = 0
    while i < count:
        wave_number = midi.le14bits(data[j + 0], data[j + 1])
        loop_number = midi.le14bits(data[j + 2], data[j + 3])
        kind = data[j + 4]
        start = midi.le21bits(data[j + 5], data[j + 6], data[j + 7])
        end = midi.le21bits(data[j + 8], data[j + 9], data[j + 10])
        loop_point = LoopPoint(
            wave_number=wave_number,
            loop_number=loop_number,
            kind=kind,
            start=start,
            end=end)
        loop_points[i] = loop_point
        j += 11
    return DumpLoopPointsTransmit(
        channel=channel,
        loop_points=loop_points)


def dump_loop_points(event):
    """ Decoded Dump Loop Points non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.SAMPLE_LOOP_POINTS):
        raise RuntimeError
    result = None
    sub_id2 = event.data[0]
    if sub_id2 == ID2s.SAMPLE_LOOP_POINTS_REQUEST:
        result = dump_loop_points_request(event)
    elif sub_id2 == ID2s.SAMPLE_LOOP_POINTS_TRANSMIT:
        result = dump_loop_points_transmit(event)
    return result


def dump_packet_checksum(data0, data1, data2, data3, data):
    """ Dump packet 7 bits checksum. """
    # `^` is bitwise xor.
    result = data0 ^ data1
    result = result ^ data2
    result = result ^ data3
    for byt in data:
        result = result ^ byt
    if result >= 0x80:
        result -= 0x80
    return result


def dump_packet(event):
    """ Decoded Dump Packet non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.SAMPLE_DUMP_PACKET):
        raise RuntimeError
    channel = event.channel
    data = event.data
    if len(data) != 122:
        raise ValueError
    packet_number = data[0]
    wave_data = data[1:121]
    expected_checksum = data[121]
    computed_checksum = dump_packet_checksum(
        data0=0x7E,
        data1=channel,
        data2=ID1s.SAMPLE_DUMP_PACKET,
        data3=packet_number,
        data=wave_data)
    error = computed_checksum != expected_checksum
    return DumpPacket(
        channel=channel,
        packet_number=packet_number,
        data=wave_data,
        error=error)


def dump_request(event):
    """ Decoded Dump Request non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.SAMPLE_DUMP_REQUEST):
        raise RuntimeError
    channel = event.channel
    data = event.data
    if len(data) != 2:
        raise ValueError
    wave_number = midi.le14bits(data[0], data[1])
    return DumpRequest(channel=channel, wave_number=wave_number)


def general_information(event):
    """ Decoded General Information non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.GENERAL_INFORMATION):
        raise RuntimeError
    result = None
    channel = event.channel
    sub_id2 = event.data[0]
    data = event.data
    if sub_id2 == ID2s.IDENTITY_REQUEST:
        if len(data) != 1:
            raise ValueError
        result = IdentityRequest(channel=channel)
    elif sub_id2 == ID2s.IDENTITY_REPLY:
        manufacturer_data = manufacturers.decoded_data(data)
        manufacturer = manufacturer_data.manufacturer
        data = manufacturer_data.data
        if len(data) != 9:
            raise ValueError
        familly = midi.le14bits(data[1], data[2])
        model = midi.le14bits(data[3], data[4])
        version = midi.le28bits(data[5], data[6], data[7], data[8])
        result = IdentityReply(
            channel=channel,
            manufacturer=manufacturer,
            familly=familly,
            model=model,
            version=version)
    return result


def gm_mode(event):
    """ Decoded GM Enable non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.GM_MODE):
        raise RuntimeError
    if len(event.data) != 1:
        raise ValueError
    channel = event.channel
    mode = event.data[0]
    if mode == 0x00:
        level = 1
        is_on = False
    elif mode == 0x01:
        level = 1
        is_on = True
    elif mode == 0x02:
        level = 2
        is_on = False
    elif mode == 0x03:
        level = 2
        is_on = True
    else:
        raise ValueError
    return GMMode(channel=channel, level=level, on=is_on)


def handshake_ack(event):
    """ Decoded ACK non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.HANDSHAKE_ACK):
        raise RuntimeError
    if len(event.data) != 1:
        raise ValueError
    channel = event.channel
    packet_number = event.data[0]
    return Handshake(
        channel=channel,
        packet_number=packet_number,
        status=Handshakes.ACK)


def handshake_cancel(event):
    """ Decoded Handshake Cancel non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.HANDSHAKE_CANCEL):
        raise RuntimeError
    if len(event.data) != 1:
        raise ValueError
    channel = event.channel
    packet_number = event.data[0]
    return Handshake(
        channel=channel,
        packet_number=packet_number,
        status=Handshakes.CANCEL)


def handshake_nak(event):
    """ Decoded Handshake NAK non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.HANDSHAKE_NAK):
        raise RuntimeError
    if len(event.data) != 1:
        raise ValueError
    channel = event.channel
    packet_number = event.data[0]
    return Handshake(
        channel=channel,
        packet_number=packet_number,
        status=Handshakes.NAK)


def handshake_wait(event):
    """ Decoded Handshake Wait non‑realtime universal SysEx. """
    if not is_of_expected_type(
            event=event,
            realtime=False,
            sub_id=ID1s.HANDSHAKE_WAIT):
        raise RuntimeError
    if len(event.data) != 1:
        raise ValueError
    channel = event.channel
    packet_number = event.data[0]
    return Handshake(
        channel=channel,
        packet_number=packet_number,
        status=Handshakes.WAIT)


# Main
# ----------------------------------------------------------------------------

NON_REALTIME_METHODS = {
    ID1s.GENERAL_INFORMATION: general_information,
    ID1s.GM_MODE: gm_mode,
    ID1s.HANDSHAKE_ACK: handshake_ack,
    ID1s.HANDSHAKE_CANCEL: handshake_cancel,
    ID1s.HANDSHAKE_NAK: handshake_nak,
    ID1s.HANDSHAKE_WAIT: handshake_wait,
    ID1s.SAMPLE_DUMP_HEADER: dump_header,
    ID1s.SAMPLE_DUMP_PACKET: dump_packet,
    ID1s.SAMPLE_DUMP_REQUEST: dump_request,
    ID1s.SAMPLE_LOOP_POINTS: dump_loop_points,
}


def non_realtime_decoded(event):
    """ Decoded non‑realtime universal SysEx. """
    if not isinstance(event, sysex_events.Universal):
        raise RuntimeError
    if event.realtime is not False:
        raise RuntimeError
    result = None
    sub_id = event.sub_id
    if sub_id in NON_REALTIME_METHODS:
        method = NON_REALTIME_METHODS[sub_id]
        result = method(event)
    return result


# Realtime
# ============================================================================

def realtime_decoded(event):
    """ Decoded realtime universal SysEx. """
    if not isinstance(event, sysex_events.Universal):
        raise RuntimeError
    if event.realtime is not True:
        raise RuntimeError
    channel = event.channel
    sub_id = event.sub_id
    sub_id2 = event.data[0]
    data = event.data
    result = None
    if sub_id == ID1s.DEVICE_CONTROL:
        if sub_id2 == ID2s.MASTER_VOLUME:
            if len(data) != 3:
                raise ValueError
            volume = midi.le14bits(data[1], data[2])
            result = MasterVolume(channel=channel, volume=volume)
    return result


# Main
# ============================================================================

def decoded(event):
    """ Decoded universal SysEx. """
    if not isinstance(event, sysex_events.Universal):
        raise RuntimeError
    if event.realtime is False:
        result = non_realtime_decoded(event)
    else:
        result = realtime_decoded(event)
    if result is None:
        result = event
    return result
