# -*- coding:utf-8; mode:python3; indent-tabs-mode:nil; c-basic-offset:4; -*-

""" MIDI Channel Voice Events. """

from collections import namedtuple

from . import channel_events
from . import midi

from .channel_events import Statuses


# Constants
# ============================================================================

DEFAULT_VELOCITY = 64  # Mezzo‑forte (or mezzo‑piano?)
PITCH_BEND_MIDDLE = 8192  # 64 * 128, 0x2000


# Types
# ============================================================================

ChannelPressure = namedtuple("ChannelPressure", ["channel", "value"])
# May have different effect, not necessarily or just volume.
# Applies to all notes.

ControlChange = namedtuple("ControlChange", ["channel", "controller", "value"])
# May be further decoded by `controller_events.decoded`.
# Does not include mode events, which are distinguished earlier.

KeyPressure = namedtuple("KeyPressure", ["channel", "key", "value"])
# May have different effect, not necessarily or just volume.
# Applies to each notes, individually.

Note = namedtuple("Note", ["channel", "key", "on", "velocity"])
# All Note On, must have a corresponding Note Off.

PitchBend = namedtuple("PitchBend", ["channel", "value"])
# Value is already centered on zero (hence a signed integer).
# Sensitivity may be changed with RPN#0.

ProgramChange = namedtuple("ProgramChange", ["channel", "program"])
# Program name and familly name, may be retreived with `gm1_sound_names`.


# Per Type
# ============================================================================

def decoded_note(channel, note_on, data):
    """ `Note` from data (and channel and `note_on`). """
    if len(data) != 2:
        raise RuntimeError
    key = data[0]  # Middle C = 60 (0x3C).
    velocity = data[1]  # Logarithmic scale.
    if note_on and velocity == 0:
        note_on = False  # May be handy with running status ;-)
        velocity = DEFAULT_VELOCITY
    return Note(channel=channel, key=key, on=note_on, velocity=velocity)


def decoded_key_pressure(channel, data):
    """ `KeyPressure` from data (and channel). """
    if len(data) != 2:
        raise RuntimeError
    key = data[0]
    value = data[1]
    return KeyPressure(channel=channel, key=key, value=value)


def decoded_control_change(channel, data):
    """ `ControlChange` from data (and channel). """
    if len(data) != 2:
        raise RuntimeError
    controller = data[0]
    if controller >= channel_events.MODE_CONTROLLER_FIRST:
        raise RuntimeError  # Handled elsewhere as channel mode.
    value = data[1]
    return ControlChange(
        channel=channel,
        controller=controller,
        value=value)


def decoded_program_change(channel, data):
    """ `ProgramChange` from data (and channel). """
    if len(data) != 1:
        raise RuntimeError
    program = data[0]
    return ProgramChange(channel=channel, program=program)


def decoded_channel_pressure(channel, data):
    """ `ChannelPressure` from data (and channel). """
    if len(data) != 1:
        raise RuntimeError
    value = data[0]
    return ChannelPressure(channel=channel, value=value)


def decoded_pitch_bend(channel, data):
    """ `PitchBend` from data (and channel). """
    if len(data) != 2:
        raise RuntimeError
    value = midi.le14bits(data[0], data[1]) - PITCH_BEND_MIDDLE
    # Increment is determined by the receiver
    return PitchBend(channel=channel, value=value)


# Main
# ============================================================================

def decoded(event):
    """ One of channel voice events. """
    if not isinstance(event, midi.VoiceEvent):
        raise RuntimeError
    opcode = (event.status // 16) * 16
    channel = (event.status % 16)
    data = event.data
    if opcode in {Statuses.NOTE_OFF, Statuses.NOTE_ON}:
        note_on = opcode == Statuses.NOTE_ON
        result = decoded_note(channel, note_on, data)
    elif opcode == Statuses.KEY_PRESSURE:
        result = decoded_key_pressure(channel, data)
    elif opcode == Statuses.CONTROL_CHANGE:
        result = decoded_control_change(channel, data)
    elif opcode == Statuses.PROGRAM_CHANGE:
        result = decoded_program_change(channel, data)
    elif opcode == Statuses.CHANNEL_PRESSURE:
        result = decoded_channel_pressure(channel, data)
    elif opcode == Statuses.PITCH_BEND:
        result = decoded_pitch_bend(channel, data)
    else:
        raise RuntimeError
    return result
